/*  sound.c
 *
 *  Sound-card handling functions of xwxapt application
 */

/*
 *  xwxapt: An application to decode APT signals from weather
 *  satellites and produce an image of the weather.
 *
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "sound.h"
#include "shared.h"

static int dsp_rate_err;  /* dsp sampling rate error*/
static guint dsp_buf_idx; /* Index for this buffer  */

/* ALSA pcm capture and mixer handles */
static snd_pcm_t *capture_handle = NULL;
static snd_mixer_t *mixer_handle = NULL;
static snd_pcm_hw_params_t *hw_params = NULL;

/*------------------------------------------------------------------------*/

/*  Setup_Sound_Card()
 *
 *  Sets up mixer and DSP device and adjusts input
 *  level for correct sampling of APT signal
 */

  gboolean
Setup_Sound_Card( char *mesg, int *error )
{
  snd_mixer_elem_t *elem = NULL;
  snd_mixer_selem_id_t *sid = NULL;;
  long cmin, cmax;


  /* Return if sound card is ready */
  if( isFlagSet(SOUND_CARD_SETUP) )	return( TRUE );

  /*** Open/setup dsp/pcm and mixer ***/
  Show_Message( _("Setting up the sound card ... "), "black" );

  /*** Set up pcm parameters ***/
  /* Open pcm for capture */
  Show_Message( _("Opening audio device for capture ... "), "black" );
  *error = snd_pcm_open(
	  &capture_handle, rc_data.snd_card,
	  SND_PCM_STREAM_CAPTURE, SND_PCM_ASYNC );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot open audio device "), MESG_SIZE );
	Strlcat( mesg, rc_data.snd_card, MESG_SIZE );
	return( FALSE );
  }

  /* Allocate memory to hardware parameters structure */
  *error = snd_pcm_hw_params_malloc( &hw_params );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot allocate hw_params struct"), MESG_SIZE );
	return( FALSE );
  }

  /* Initialize hardware parameter structure */
  *error = snd_pcm_hw_params_any( capture_handle, hw_params );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot initialize hw_params struct"), MESG_SIZE );
	return( FALSE );
  }

  /* Set access type */
  *error = snd_pcm_hw_params_set_access(
	  capture_handle, hw_params,
	  SND_PCM_ACCESS_RW_INTERLEAVED );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot set access type"), MESG_SIZE );
	return( FALSE );
  }

  /* Set sample format */
  *error = snd_pcm_hw_params_set_format(
	  capture_handle, hw_params,
	  SND_PCM_FORMAT_S16_LE );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot set sample format"), MESG_SIZE );
	return( FALSE );
  }

  /* Set sample rate */
  *error = snd_pcm_hw_params_set_rate(
	  capture_handle, hw_params, DSP_RATE, EXACT_VAL );
  if( *error < 0 )
  {
	snprintf( mesg, MESG_SIZE,
		_("Cannot set sample rate to %d"), DSP_RATE );
	return( FALSE );
  }

  /* Set channel count */
  *error = snd_pcm_hw_params_set_channels(
	  capture_handle, hw_params, rc_data.num_chn );
  if( *error < 0 )
  {
	snprintf( mesg, MESG_SIZE,
		_("Cannot set channel count to %d"), rc_data.num_chn );
	return( FALSE );
  }

  /* Set number of periods */
  *error = snd_pcm_hw_params_set_periods(
	  capture_handle, hw_params, NUM_PERIODS, EXACT_VAL );
  if( *error < 0)
  {
	snprintf( mesg, MESG_SIZE,
		_("Cannot set number periods to %d"), NUM_PERIODS );
	return( FALSE );
  }

  /* Set period size */
  *error = snd_pcm_hw_params_set_period_size(
	  capture_handle, hw_params, READI_FRAMES, EXACT_VAL);
  if( *error < 0)
  {
	snprintf( mesg, MESG_SIZE,
		_("Cannot set period size to %d"), READI_FRAMES );
	return( FALSE );
  }

  /* Set parameters */
  *error = snd_pcm_hw_params( capture_handle, hw_params );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot set capture parameters"), MESG_SIZE );
	return( FALSE );
  }
  snd_pcm_hw_params_free( hw_params );
  hw_params = NULL;

  /*** Set up mixer (capture source and volume) ***/
  /* Open mixer */
  Show_Message( _("Opening mixer device ... "), "black" );
  *error = snd_mixer_open( &mixer_handle, 0 );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot open mixer"), MESG_SIZE );
	return( FALSE );
  }

  /* Attach mixer */
  *error = snd_mixer_attach( mixer_handle, rc_data.snd_card );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot attach mixer to "), MESG_SIZE );
	Strlcat( mesg, rc_data.snd_card, MESG_SIZE );
	return( FALSE );
  }

  /* Register mixer */
  *error = snd_mixer_selem_register( mixer_handle, NULL, NULL );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot register mixer"), MESG_SIZE );
	return *error;
  }

  /* Load mixer */
  *error = snd_mixer_load( mixer_handle );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot load mixer"), MESG_SIZE );
	return( FALSE );
  }

  /* Allocate selem_id structure */
  *error = snd_mixer_selem_id_malloc( &sid );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot allocate selem_id struct"), MESG_SIZE );
	return( FALSE );
  }

  /* Find capture selem */
  if( strcmp(rc_data.cap_src, "--") != 0 )
  {
	snd_mixer_selem_id_set_index( sid, 0 );
	snd_mixer_selem_id_set_name( sid, rc_data.cap_src );
	elem = snd_mixer_find_selem( mixer_handle, sid );
	if( elem == NULL )
	{
	  Strlcat( mesg, _("Cannot find capture element "), MESG_SIZE );
	  Strlcat( mesg, rc_data.cap_src, MESG_SIZE );
	  snd_mixer_selem_id_free(sid);
	  *error = 0;
	  return( FALSE );
	}
  }
  else rc_data.channel = -1;

  /* Set capture switch for cap source */
  if( rc_data.channel != -1 )
  {
	Show_Message( _("Setting capture source ... "), "black" );
	if( snd_mixer_selem_has_capture_switch(elem) )
	{
	  *error = snd_mixer_selem_set_capture_switch(
		  elem, rc_data.channel, 1 );
	  if( *error < 0 )
	  {
		Strlcat( mesg, _("Cannot set capture device "), MESG_SIZE );
		Strlcat( mesg, rc_data.cap_src, MESG_SIZE );
		snd_mixer_selem_id_free(sid);
		return( FALSE );
	  }
	}
	else
	{
	  Strlcat( mesg, _("Element has no Capture capability: "), MESG_SIZE );
	  Strlcat( mesg, rc_data.cap_src, MESG_SIZE );
	  snd_mixer_selem_id_free(sid);
	  *error = 0;
	  return( FALSE );
	}
  }

  /* Find capture volume selem */
  if( strcmp(rc_data.cap_vol, "--") != 0 )
  {
	snd_mixer_selem_id_set_index( sid, 0 );
	snd_mixer_selem_id_set_name( sid, rc_data.cap_vol );
	elem = snd_mixer_find_selem( mixer_handle, sid );
	if( elem == NULL )
	{
	  Strlcat( mesg, _("Cannot find volume element "), MESG_SIZE );
	  Strlcat( mesg, rc_data.cap_vol, MESG_SIZE );
	  Show_Message( mesg, "red" );
	}

	/* Set capture volume */
	if( elem != NULL )
	{
	  if( snd_mixer_selem_has_capture_volume(elem) )
	  {
		/* Change from % volume to sound card value */
		long lev;
		snd_mixer_selem_get_capture_volume_range( elem, &cmin, &cmax );
		lev = cmin + ((cmax - cmin) * rc_data.cap_lev) / 100;

		/* Set capture volume */
		Show_Message( _("Setting capture volume ... "), "black" );
		*error = snd_mixer_selem_set_capture_volume(
			elem, rc_data.channel, lev );
		if( *error < 0 )
		{
		  snprintf( mesg, MESG_SIZE,
			  _("Cannot set capture volume to %d"), rc_data.cap_lev );
		  Show_Message( mesg, "red" );
		}
	  }
	  else
	  {
		Strlcat( mesg, _("Element has no Volume capability: "), MESG_SIZE );
		Strlcat( mesg, rc_data.cap_src, MESG_SIZE );
		snd_mixer_selem_id_free(sid);
		*error = 0;
		return( FALSE );
	  }
	} /* if( elem != NULL ) */
  } /* if( strcmp(rc_data.cap_vol, "--") != 0 ) */

  snd_mixer_selem_id_free(sid);
  Close_Mixer_Handle();

  /* Allocate memory to DSP data buffer */
  rc_data.dsp_buf_size = rc_data.num_chn * READI_FRAMES;
  if( gbl_dsp_buffer == NULL )
	gbl_dsp_buffer = malloc( (size_t)rc_data.dsp_buf_size * sizeof(short) );
  if( gbl_dsp_buffer == NULL )
  {
	Show_Message( _("malloc() for samples buffer failed "), "red" );
	*error = 0;
	Error_Dialog();
	return( FALSE );
  }
  memset( gbl_dsp_buffer, 0,
	  (size_t)rc_data.dsp_buf_size * sizeof(short) );

  /* This will make Read_Dsp_Buffer() initialize */
  dsp_buf_idx = rc_data.dsp_buf_size;

  /* Prepare audio interface for use */
  Show_Message( _("Preparing audio interface for use ... "), "black" );
  *error = snd_pcm_prepare( capture_handle );
  if( *error < 0 )
  {
	Strlcat( mesg, _("Cannot prepare audio interface"), MESG_SIZE );
	return( FALSE );
  }

  /* Succeeded, show a message and set flag */
  Show_Message( _("Sound card set-up completed"), "green" );
  SetFlag( SOUND_CARD_SETUP );

  return( TRUE );
} /* End of Setup_Sound_Card() */

/*------------------------------------------------------------------------*/

/* These two functions close soundcard handles */
  void
Close_Capture_Handle( void )
{
  if( capture_handle != NULL )
	snd_pcm_close( capture_handle );
  capture_handle = NULL;

  if( hw_params != NULL )
	snd_pcm_hw_params_free( hw_params );
  hw_params = NULL;

  ClearFlag( SOUND_CARD_SETUP );
}

  void
Close_Mixer_Handle( void )
{
  if( mixer_handle != NULL )
	snd_mixer_close( mixer_handle );
  mixer_handle = NULL;

  ClearFlag( SOUND_CARD_SETUP );
}

/*------------------------------------------------------------------------*/

/*  Setup_Audio_Level()
 *
 *  Displays a simple graphical aid for setting
 *  up the audio level from the APT receiver
 */

  gboolean
Setup_Audio_Level( gpointer data )
{
  int
	carrier_max, /* Max value of carrier in  buffer */
	carrier_val, /* Sampled value of 2.4kHz carrier */
	idx;         /* Index used in loops etc */

  /* Progressbar fraction */
  double pbar_frac;

  /* Required max value of carrier level */
  static int level_max;

  /* First idle call */
  static gboolean first_call = TRUE;

  /* Setup sound card, abort on error */
  if( first_call )
  {
	char mesg[MESG_SIZE];
	mesg[0] = '\0';
	int error;
	if( !Setup_Sound_Card(mesg, &error) )
	{
	  if( error )
	  {
		Strlcat( mesg, _("\nError: "), MESG_SIZE );
		Strlcat( mesg, snd_strerror(error), MESG_SIZE );
	  }

	  Cleanup();
	  Show_Message( mesg, "red" );
	  Error_Dialog();
	  first_call = TRUE;
	  return( FALSE );
	}

	/* Select required max audio level */
	switch( gbl_sat_type )
	{
	  case NOAA:
		level_max = NOAA_CARRIER_MAX;
		break;

	  case METEOR:
		level_max = METEOR_CARRIER_MAX;
	}

	Show_Message( _("Starting Audio Level Setup ..."), "black" );
	first_call = FALSE;
  }

  /* Fill in the samples buffer from DSP, abort on error */
  if( !Read_Dsp_Buffer(gbl_line_buffer, BLOCK_SIZE) )
  {
	first_call = TRUE;
	return( FALSE );
  }

  /*** Find max value of 2.4kHz subcarrier in buffer ***/
  carrier_max = 0;
  for( idx = 0; idx < BLOCK_SIZE-1; idx++ )
  {
	carrier_val = Carrier_Amplitude(
		gbl_line_buffer[idx],
		gbl_line_buffer[idx+1]);

	/* Find maximum carrier level */
	if( carrier_val > carrier_max )
	  carrier_max = carrier_val;
  }

  /* Show audio level on progress bar */
  pbar_frac = (gdouble)carrier_max / (gdouble)level_max ;
  if( pbar_frac > 1.0 )	pbar_frac = 1.0;
  gtk_progress_bar_set_fraction(
	  GTK_PROGRESS_BAR(gbl_level_pbar), pbar_frac );

  if( isFlagSet(ACTION_SETUP_AUDIO) )
	return( TRUE );

  Show_Message( _("Stopping Audio Level Setup"), "black" );
  Cleanup();
  first_call = TRUE;

  return( FALSE );
} /* Setup_Audio_Level() */

/*------------------------------------------------------------------------*/

/*  Synchronize_Buffer()
 *
 *  Synchronizes samples buffer with the APT sync train
 */

  gboolean
Synchronize_Buffer( gpointer data )
{
  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* Messages buffer */
  char mesg[MESG_SIZE];

  /* Data block size used as read() argument  */
  int block_size;

  static int
	last_line_idx,		/* Saved value of samples buffer index */
	rate_ok_cnt;		/* Count number of similar rate measurements */

  /* Cancel on user request */
  if( isFlagClear(ACTION_SYNC_BUFFER) )
  {
	Show_Message( _("Search for APT sync train stopped"), "black" );
	Cleanup();
	first_call = TRUE;
	return( FALSE );
  }

  /* Initialize on first call */
  if( first_call )
  {
	int error;
	mesg[0] = '\0';
	if( !Setup_Sound_Card(mesg, &error) )
	{
	  if( error )
	  {
		Strlcat( mesg, _("\nError: "), MESG_SIZE );
		Strlcat( mesg, snd_strerror(error), MESG_SIZE );
	  }

	  Cleanup();
	  Show_Message( mesg, "red" );
	  Error_Dialog();
	  first_call = TRUE;
	  return( FALSE );
	}

	/* Print some information */
	Show_Message( _("Synchronizing with APT sync train ..."), "black" );
	snprintf( mesg, MESG_SIZE,
		_("Looking for a %s type sync train ... "),	gbl_sat_names[gbl_sat_type] );
	Show_Message( mesg, "black" );

	/* Load appropriate sync reference */
	gbl_sync_ref = gbl_sync_refs[ gbl_sat_type ];

	dsp_rate_err  = 0;
	last_line_idx = 0;
	rate_ok_cnt   = 0;

	first_call = FALSE;
	SetFlag(ACTION_RATE_ERROR);

  } /* if( first_call ) */

  /*** Find DSP rate error ***/
  if( isFlagSet(ACTION_RATE_ERROR) )
  {
	/* Fill in the line buffer from DSP */
	if( !Read_Dsp_Buffer(gbl_line_buffer, BLOCK_SIZE) )
	{
	  first_call = TRUE;
	  return( FALSE );
	}

	/* Locate sync train (sets gbl_line_idx just after it) */
	if( !Detect_Sync_Train(BLOCK_SIZE) ) return( TRUE );

	/* DSP rate error is change in sync train position */
	dsp_rate_err = gbl_line_idx - last_line_idx;
	last_line_idx = gbl_line_idx;

	/* If measured dsp sampling rate error is
	 * within limits, signal end of measurement */
	if( abs(dsp_rate_err) > SYNC_ACCEPT )
	  rate_ok_cnt = 0;
	else
	  rate_ok_cnt++;
	if( rate_ok_cnt >= RATE_OK_COUNT )
	  ClearFlag(ACTION_RATE_ERROR);
	else return( TRUE );

	/* Dummy-read samples to align buffer with sync ref. */
	block_size = gbl_line_idx - gbl_sync_ref;
	if( block_size < 0 )
	  block_size += BLOCK_SIZE;
	block_size += (dsp_rate_err * block_size)/BLOCK_SIZE;
	if( !Read_Dsp_Buffer(gbl_line_buffer, block_size) )
	{
	  first_call = TRUE;
	  return( FALSE );
	}

	/* Show dsp rate error */
	snprintf( mesg, MESG_SIZE,
		_("DSP sampling rate error: %d ppm"),
		(1000000 * dsp_rate_err) / BLOCK_SIZE );
	Show_Message( mesg, "orange" );
	gbl_sync_err = 0;

  } /* if( isFlagSet(ACTION_RATE_ERROR) ) */

  /*** Refine buffer alignment with sync reference ***/
  /* Dummy-read samples to align buffer with sync reference */
  block_size = BLOCK_SIZE + dsp_rate_err + gbl_sync_err;
  if( !Read_Dsp_Buffer(gbl_line_buffer, block_size) )
  {
	first_call = TRUE;
	return( FALSE );
  }

  /* Locate sync train. (sets gbl_line_idx just after it) */
  if( !Detect_Sync_Train(BLOCK_SIZE) ||
	  abs(gbl_sync_err) > SYNC_ACCEPT )
	return( TRUE );

  /* Print some information */
  Show_Message( _("Synchronized with APT sync train"), "green" );
  first_call = TRUE;
  ClearFlag(ACTION_SYNC_BUFFER);

  /* Start the selected process */
  if( isFlagSet(ACTION_PROCESS_DSP) )
	gtk_idle_add( Dsp_Process_Image, NULL );
  else if( isFlagSet(ACTION_RECORD_APT) )
	gtk_idle_add( Record_APT_Signal, NULL );

  return( FALSE );
} /* End of Synchronize_Buffer() */

/*------------------------------------------------------------------------*/

/*  Record_APT_Signal()
 *
 *  Records samples of the 2.4 Khz subcarrier in a file
 */

  gboolean
Record_APT_Signal( gpointer data )
{
  /* Number of APT image lines processed */
  static int line_cnt;

  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* Messages buffer */
  char mesg[256];
  size_t sm = sizeof( mesg );

  /* Temp buffer to write line buffer to file */
  unsigned char buff[2];
  int idx;

  /* Samples written to file */
  size_t sample_cnt;


  if( first_call )
  {
	Show_Message( _("Recording apt signal to file ..."), "black" );

	/* Add extension to file name */
	size_t s = sizeof( gbl_samples_file );
	Strlcat( gbl_samples_file, "-", s );
	Strlcat( gbl_samples_file, gbl_sat_names[gbl_sat_type], s );
	Strlcat( gbl_samples_file, ".bin", s );

	/* Create a file for writing samples */
	gbl_samples_fp = fopen( gbl_samples_file, "w+" );
	if( gbl_samples_fp == NULL )
	{
	  perror( gbl_samples_file );
	  snprintf( mesg, sm,
		  _("Failed to open recorded samples file:\n%s"),
		  gbl_samples_file );
	  Show_Message( mesg, "red" );
	  Error_Dialog();
	  return( FALSE );
	}

	/* Read samples from DSP, stop on user interrupt */
	/* or at time limit in number of APT lines read. */
	/* gbl_duration is in seconds of recording time. */
	snprintf( mesg, sm,
		_("Setting recording duration to %d sec"), gbl_duration );
	Show_Message( mesg, "black" );

	/* Print some information */
	snprintf( mesg, sm,
		_("Recording APT signal in:\n%s ..."), gbl_samples_file );
	Show_Message( mesg, "black" );

	line_cnt = 0;

	/* Write satellite type to samples file, abort on error */
	sample_cnt = fwrite(&gbl_sat_type, 4, 1, gbl_samples_fp);
	if( sample_cnt < 1 )
	{
	  perror( "xwxapt: fwrite()" );
	  snprintf( mesg, sm,
		  _("Error writing to samples file\n%s"), gbl_samples_file );
	  Show_Message( mesg, "red" );
	  Error_Dialog();
	  return( FALSE );
	}
	if( sample_cnt != 1 )
	{
	  snprintf( mesg, sm,
		  _("Error writing to samples file\n%s"), gbl_samples_file );
	  Show_Message( mesg, "red" );
	  Error_Dialog();
	  return( FALSE );
	}

	first_call = FALSE;

  } /* if( first_call ) */

  /* Fill samples buffer from DSP */
  if( !Dsp_Fill_Buffer() )
  {
	first_call = TRUE;
	return( FALSE );
  }

  /* Write carrier sample values to samples file, abort on error */
  for( idx = 0; idx < BLOCK_SIZE; idx++ )
  {
	buff[0] = gbl_line_buffer[idx] & 0xff;
	buff[1] = (gbl_line_buffer[idx] >> 8) & 0xff;

	sample_cnt = fwrite(buff, 1, 2, gbl_samples_fp);
	if( sample_cnt < 1 )
	{
	  perror( "xwxapt: fwrite()" );
	  snprintf(mesg, sm,
		  _("Error writing to samples file\n%s"), gbl_samples_file);
	  Show_Message( mesg, "red" );
	  Error_Dialog();
	  first_call = TRUE;
	  return( FALSE );
	}

	if( (sample_cnt != 2) )
	{
	  snprintf(mesg, sm,
		  _("Error writing to samples file\n%s"), gbl_samples_file);
	  Show_Message( mesg, "red" );
	  Error_Dialog();
	  first_call = TRUE;
	  return( FALSE );
	}
  } /* for( idx = 0; idx < BLOCK_SIZE; idx++ ) */

  /* Stop if line count reaches time */
  /* limit or on user stop request   */
  if( (++line_cnt/2 != gbl_duration) &&
	  isFlagSet(ACTION_RECORD_APT) )
	return( TRUE );

  /* Print some information */
  snprintf(mesg, sm,
	  _("Closing samples file:\n%s"), gbl_samples_file);
  Show_Message( mesg, "black" );

  Cleanup();
  first_call = TRUE;

  return( FALSE );
} /* End of Record_APT_Signal() */

/*------------------------------------------------------------------------*/

/*  File_Fill_Buffer()
 *
 *  Fills the audio sample buffer from source file
 */

  gboolean
File_Fill_Buffer( void )
{
  /* Messages buffer */
  char mesg[256];

  /* For reading processed dsp data from file */
  size_t sample_cnt;

  /* Temp buffer to write line buffer to file */
  unsigned char buff[2];
  int idx;


  /* Fill in the samples buffer from file, return on error */
  gbl_line_idx = 0;
  for( idx = 0; idx < BLOCK_SIZE; idx++ )
  {
	sample_cnt = fread(buff, 1, 2, gbl_samples_fp);
	if( sample_cnt < 2 )
	{
	  perror( "xwxapt: fread()" );
	  snprintf(mesg, sizeof(mesg),
		  _("Error reading from samples file\n%s"), gbl_samples_file);
	  Show_Message( mesg, "red" );
	  Error_Dialog();
	  return( FALSE );
	}

	if( feof(gbl_samples_fp) )
	  return( FALSE );
	else if( sample_cnt != 2 )
	{
	  snprintf(mesg, sizeof(mesg),
		  _("Error reading from samples file\n%s"), gbl_samples_file);
	  Show_Message( mesg, "red" );
	  Error_Dialog();
	  return( FALSE );
	}

	gbl_line_buffer[idx]  = buff[0];
	gbl_line_buffer[idx] |= buff[1] << 8;
	if( gbl_line_buffer[idx] & 0x8000 )
	  gbl_line_buffer[idx] -= 0x10000;

  } /* for( idx = 0; idx < BLOCK_SIZE; idx++ ) */

  /* Set buffer index just after sync train */
  /* Search for sync up to 1.5* sync refer. */
  if( !Detect_Sync_Train((3*gbl_sync_refs[gbl_sat_type])/2) )
	return( TRUE );

  /* If sync detection is far out, keep */
  /* buffer index in default position   */
  if( abs(gbl_line_idx - gbl_sync_ref) > SYNC_ACCEPT )
	gbl_line_idx = gbl_sync_ref;

  return( TRUE );
} /* End of File_Fill_Buffer() */

/*------------------------------------------------------------------------*/

/*  Dsp_Fill_Buffer()
 *
 *  Fills the audio sample buffer from sound card DSP.
 *  Buffer filling is synchronized with APT sync train.
 */

  gboolean
Dsp_Fill_Buffer( void )
{
  static int block_size = BLOCK_SIZE;

  /* Fill in the samples buffer from DSP */
  if( !Read_Dsp_Buffer(gbl_line_buffer, block_size + dsp_rate_err) )
  {
	Show_Message( _("Error filling samples buffer from dsp"), "red" );
	return( FALSE );
  }

  /* Search for sync up to 1.5 * sync ref, return if sync fails */
  if( !Detect_Sync_Train((3 * gbl_sync_ref) / 2) )
  {
	gbl_line_idx = gbl_sync_ref;
	return( TRUE );
  }

  /* Try to keep APT sync train at beginning of buffer. This  */
  /* done done by increasing or decreasing the data  block    */
  /* size argument passed to read(), by adding or subtracting */
  /* half the difference between detected sync position and a */
  /* reference value. Error is limited to +-2 to avoid coarse */
  /* corrections to sync position due to noise interference.  */
  if( gbl_sync_err > SYNC_ACCEPT )
	gbl_sync_err = SYNC_ACCEPT;
  else
	if( gbl_sync_err < -SYNC_ACCEPT )
	  gbl_sync_err = -SYNC_ACCEPT;
  gbl_line_idx += gbl_sync_err;

  /* Vary the data block size argument to read() to
   * take up any drift in sync position due to errors
   * in the DSP sampling rate and satellite movement.
   * Only half the sync error is used to slow and
   * stabilize sync correction. */
  block_size = BLOCK_SIZE + gbl_sync_err/2;

  return( TRUE );
} /* End of Dsp_Fill_Buffer() */

/*------------------------------------------------------------------------*/

/* Read_Dsp_Buffer()
 *
 * Reads the DSP's buffer and separates Left/Right
 * channel data depending on stereo or mono mode
 */

  gboolean
Read_Dsp_Buffer( short *buffer, int buff_size )
{
  int idx;
  snd_pcm_sframes_t error;

  /* Transfer data from dsp buffer to buffer to be processed, */
  /* separating left or right channel data if in stereo mode  */
  if( buff_size > BUFFER_SIZE ) buff_size -= BUFFER_SIZE;
  for( idx = 0; idx < buff_size; idx++ )
  {
	if( dsp_buf_idx >= rc_data.dsp_buf_size )
	{
	  /* Start buffer index according to channel */
	  dsp_buf_idx = rc_data.use_chn;

	  /* Fill in the samples buffer from DSP, return -1 on error */
	  error = snd_pcm_readi(
		  capture_handle, gbl_dsp_buffer, READI_FRAMES );
	  if( error != READI_FRAMES )
	  {
		char mesg[MESG_SIZE];
		snprintf( mesg, MESG_SIZE,
			_("Read from audio interface failed\n"\
			  "Error: %s"), snd_strerror((int)error) );
		Show_Message( mesg, "red" );

		/* Try to recover from error */
		if( !Xrun_Recovery((int)error) )
		{
		  snprintf( mesg, MESG_SIZE,
			  _("Failed to recover from %s"), snd_strerror((int)error) );
		  Show_Message( mesg, "red" );
		  Cleanup();
		  Error_Dialog();
		  return( FALSE );
		}
	  } /* if( error != NUM_FRAMES ) */
	} /* if( dsp_buf_idx >= rc_data.dsp_buf_size ) */

	/* Fill buffer from left or right channel. dsp_buf_idx
	 * is arranged to point to the correct location */
	buffer[idx] = gbl_dsp_buffer[dsp_buf_idx];

	/* Increment according to mono/stereo mode */
	dsp_buf_idx += rc_data.num_chn;

  } /* for( idx = 0; idx < buff_size; idx++ ) */

  return( TRUE );
} /* End of Read_Dsp_Buffer() */

/*------------------------------------------------------------------------*/

/* Xrun_Recovery()
 *
 * Recover from underrrun (broken pipe) and suspend
 */
  gboolean
Xrun_Recovery( int error )
{
  if( error == -EPIPE )
  {
	error = snd_pcm_prepare( capture_handle );
	if( error < 0 )
	{
	  char mesg[MESG_SIZE];
	  snprintf( mesg, MESG_SIZE,
		  _("Cannot recover from underrun, prepare failed\n"\
			"Error: %s\n"), snd_strerror(error) );
	  Show_Message( mesg, "red" );
	  return( FALSE );
	}
  }
  else if( error == -ESTRPIPE )
  {
	while( (error = snd_pcm_resume(capture_handle)) == -EAGAIN )
	  sleep(1);
	if( error < 0 )
	{
	  error = snd_pcm_prepare( capture_handle );
	  if( error < 0 )
	  {
		char mesg[MESG_SIZE];
		snprintf( mesg, MESG_SIZE,
			_("Cannot recover from suspend, prepare failed\n"\
			  "Error: %s\n"), snd_strerror(error) );
		Show_Message( mesg, "red" );
		return( FALSE );
	  }
	}
  }

  return( TRUE );
} /* Xrun_Recovery() */

/*------------------------------------------------------------------------*/

